@@ -28,7 +28,8 @@ import qualified Data.Text as T
 import qualified Data.Text.Encoding as TE
 import qualified Data.Vector as V
 import GHC.Generics (Generic)
-import System.FilePath (replaceFileName, takeBaseName, takeDirectory, (</>))
+import System.FilePath (replaceFileName, takeBaseName,
+                        takeFileName, takeDirectory, (</>))
 import Text.Parsec
 import Typst.Bind (destructuringBind)
 import Typst.Constructors (getConstructor)
@@ -100,6 +101,10 @@ satisfyTok f = tokenPrim show showPos match'
 pContent :: Monad m => MP m (Seq Content)
 pContent = (pTxt <|> pElt) >>= applyShowRules >>= addTextElement
 
+applyShowRulesToVal :: Monad m => Val -> MP m Val
+applyShowRulesToVal (VContent cs) = VContent <$> applyShowRules cs
+applyShowRulesToVal x = pure x
+
 addTextElement :: Monad m => Seq Content -> MP m (Seq Content)
 addTextElement = foldM go mempty
   where
@@ -427,7 +432,7 @@ pExpr :: Monad m => Expr -> MP m (Seq Content)
 pExpr expr = valToContent <$> evalExpr expr
 
 evalExpr :: Monad m => Expr -> MP m Val
-evalExpr expr =
+evalExpr expr = applyShowRulesToVal =<<
   case expr of
     Literal lit -> pure $ evalLiteral lit
     Group e -> evalExpr e
@@ -685,6 +690,8 @@ evalExpr expr =
       case (v1, v2) of
         (VAlignment x1 y1, VAlignment x2 y2) ->
           pure $ VAlignment (x1 `mplus` x2) (y1 `mplus` y2)
+        (VArguments xs, VArguments ys) ->
+          pure $ VArguments (xs <> ys)
         _ -> case maybePlus v1 v2 of
           Nothing -> fail $ "Can't + " <> show v1 <> " and " <> show v2
           Just v -> pure v
@@ -742,6 +749,7 @@ evalExpr expr =
                 st
                   { evalShowRules =
                       ShowRule
+                        (evalNextShowRuleIdentifier st)
                         selector
                         ( \c ->
                             valToContent
@@ -751,12 +759,14 @@ evalExpr expr =
                                     named = OM.empty
                                   }
                         )
-                        : evalShowRules st
+                        : evalShowRules st,
+                    evalNextShowRuleIdentifier = evalNextShowRuleIdentifier st + 1
                   }
             _ -> updateState $ \st ->
               st
                 { evalShowRules =
                     ShowRule
+                      (evalNextShowRuleIdentifier st)
                       selector
                       ( \c ->
                           case e of
@@ -764,7 +774,8 @@ evalExpr expr =
                             Set _ _ -> pure $ Seq.singleton c
                             _ -> pure (valToContent renderVal)
                       )
-                      : evalShowRules st
+                      : evalShowRules st,
+                    evalNextShowRuleIdentifier = evalNextShowRuleIdentifier st + 1
                 }
           pure VNone
     Binding _ -> fail $ "Encountered binding out of proper context"
@@ -994,28 +1005,35 @@ findPackageEntryPoint modname = do
 loadModule :: Monad m => Text
            -> MP m (Seq Content, (Identifier, M.Map Identifier Val))
 loadModule modname = do
-  (fp, modid, mbPackageRoot) <-
+  (fp, modid, mbPackageRoot, txt) <-
         if T.take 1 modname == "@"
            then do
             fp' <- findPackageEntryPoint modname
-            pure (fp',
+            operations <- evalOperations <$> getState
+            txt <- lift $ TE.decodeUtf8 <$> loadBytes operations fp'
+            pure ( takeFileName fp',
                    Identifier
                     (T.pack $ takeWhile (/= ':') . takeBaseName $
                        T.unpack modname),
-                   Just (takeDirectory fp'))
-           else pure (T.unpack modname,
-                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
-                        Nothing)
-  txt <- loadFileText fp
+                   Just (takeDirectory fp'),
+                   txt )
+           else do
+             let fp = T.unpack modname
+             txt <- loadFileText fp
+             pure ( fp,
+                    Identifier (T.pack $ takeBaseName fp),
+                    Nothing,
+                    txt )
   case parseTypst fp txt of
     Left err -> fail $ show err
     Right ms -> do
       operations <- evalOperations <$> getState
       currentLocalDir <- evalLocalDir <$> getState
+      currentPackageRoot <- evalPackageRoot <$> getState
       let (pkgroot , localdir) =
             case mbPackageRoot of
               Just r -> (r , takeDirectory fp)
-              Nothing -> (evalPackageRoot initialEvalState ,
+              Nothing -> (currentPackageRoot,
                           currentLocalDir </> takeDirectory fp)
       res <-
         lift $
