@@ -17,7 +17,7 @@ import Paths_typst (version)
 import Data.Version (showVersion)
 import Control.Applicative ((<|>))
 import Control.Monad (mplus, unless)
-import Control.Monad.Reader (lift)
+import Control.Monad.Reader (lift, ReaderT)
 import qualified Data.Aeson as Aeson
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.Csv as Csv
@@ -222,7 +222,7 @@ visualize :: [(Identifier, Val)]
 visualize =
   [ makeElement Nothing "circle" [("body", One (TContent :|: TNone))],
     makeElement Nothing "ellipse" [("body", One (TContent :|: TNone))],
-    makeElement Nothing "image" [("path", One TString)],
+    makeElement Nothing "image" [("source", One (TString :|: TBytes))],
     makeElement Nothing "line" [],
     makeElement Nothing "path" [("vertices", Many TArray)],
     makeElement Nothing "polygon" [("vertices", Many TArray)],
@@ -232,7 +232,7 @@ visualize =
 
 meta :: [(Identifier, Val)]
 meta =
-  [ makeElement Nothing "bibliography" [("path", One (TString :|: TArray))],
+  [ makeElement Nothing "bibliography" [("source", One (TString :|: TArray :|: TBytes))],
     makeElement Nothing "cite" [("key", One TLabel)],
     makeElement Nothing "document" [],
     makeElementWithScope
@@ -290,6 +290,7 @@ types =
   [ ("array", VType TArray)
   , ("bool", VType TBoolean)
   , ("content", VType TContent)
+  , ("dictionary", VType TDict)
   , ("int", VType TInteger)
   , ("float", VType TFloat)
   , ("regex", VType TRegex)
@@ -300,6 +301,7 @@ types =
   , ("str", VType TString)
   , ("label", VType TLabel)
   , ("version", VType TVersion)
+  , ("bytes", VType TBytes)
   ]
 
 colors :: [(Identifier, Val)]
@@ -532,16 +534,14 @@ dataLoading =
     ),
     ( "json",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case Aeson.eitherDecode bs of
           Left e -> fail e
           Right (v :: Val) -> pure v
     ),
     ( "yaml",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case Yaml.decodeEither' (BL.toStrict bs) of
           Left e -> fail $ show e
           Right (v :: Val) -> pure v
@@ -549,21 +549,23 @@ dataLoading =
     ( "read",
       makeFunction $ do
         fp <- nthArg 1
-        t <- lift $ loadFileText fp
-        pure $ VString t
+        enc <- namedArg "encoding" (VString "utf-8")
+        case enc of
+          VNone -> do bs <- lift $ loadFileLazyBytes fp
+                      pure $ VBytes $ BL.toStrict bs
+          _ -> do t <- lift $ loadFileText fp
+                  pure $ VString t
     ),
     ( "toml",
       makeFunction $ do
-        fp <- nthArg 1
-        t <- lift $ loadFileText fp
-        case Toml.decode t of
+        bs <- getFileOrBytes
+        case Toml.decode (TE.decodeUtf8 $ BL.toStrict bs) of
           Toml.Failure e -> fail (unlines ("toml errors:" : e))
           Toml.Success _ v -> pure v
     ),
     ( "xml",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case XML.parseLBS XML.def bs of
           Left e -> fail $ show e
           Right doc ->
@@ -612,3 +614,11 @@ initialEvalState =
                  , evalMathIdentifiers = [(BlockScope, mathModule <> symModule)]
                  , evalStandardIdentifiers = [(BlockScope, standardModule)]
                  }
+
+getFileOrBytes :: Monad m => ReaderT Arguments (MP m) BL.ByteString
+getFileOrBytes = do
+  v <- nthArg 1
+  case v of
+    VString fp -> lift $ loadFileLazyBytes (T.unpack fp)
+    VBytes bs -> pure $ BL.fromStrict bs
+    _ -> fail "expecting file path or bytes"
